// native.cxx -- FGFS "Native" protocal class
//
// Written by Curtis Olson, started November 1999.
//
// Copyright (C) 1999  Curtis L. Olson - curt@flightgear.org
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: native.cxx,v 1.4 2000/02/16 22:01:09 curt Exp $


#include <simgear/debug/logstream.hxx>
#include <simgear/math/fg_geodesy.hxx>

#include <Time/fg_time.hxx>

#include "iochannel.hxx"
#include "native.hxx"


FGNative::FGNative() {
}

FGNative::~FGNative() {
}


// open hailing frequencies
bool FGNative::open() {
    if ( is_enabled() ) {
	FG_LOG( FG_IO, FG_ALERT, "This shouldn't happen, but the channel " 
		<< "is already in use, ignoring" );
	return false;
    }

    FGIOChannel *io = get_io_channel();

    if ( ! io->open( get_direction() ) ) {
	FG_LOG( FG_IO, FG_ALERT, "Error opening channel communication layer." );
	return false;
    }

    set_enabled( true );

    return true;
}


// process work for this port
bool FGNative::process() {
    FGIOChannel *io = get_io_channel();
    int length = sizeof(*cur_fdm_state);

    if ( get_direction() == out ) {
	// cout << "size of cur_fdm_state = " << length << endl;
	buf = *cur_fdm_state;
	if ( ! io->write( (char *)(& buf), length ) ) {
	    FG_LOG( FG_IO, FG_ALERT, "Error writing data." );
	    return false;
	}
    } else if ( get_direction() == in ) {
	while ( io->read( (char *)(& buf), length ) == length ) {
	    FG_LOG( FG_IO, FG_ALERT, "Success reading data." );
	    *cur_fdm_state = buf;
	}
    }

    return true;
}


// close the channel
bool FGNative::close() {
    FGIOChannel *io = get_io_channel();

    set_enabled( false );

    if ( ! io->close() ) {
	return false;
    }

    return true;
}
